

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <wchar.h>
#include <locale.h>
#include <termios.h>


#define FILESIZE  1384152
#define FILELINES 22700
#define MININDEX 12293
#define MAXINDEX 65536
#define NUMINDEX ((1+MAXINDEX)-MININDEX)


typedef struct indexhead{
	int size;
	int used;
	short int *array;
} indexhead_t;


char		dict_all_lines[FILESIZE + 32];
char * const	dict_all_lines_end = dict_all_lines + FILESIZE;
char *		dict_index_lines[FILELINES + 1];
indexhead_t	dict_index_kanji[MAXINDEX + 1];


static inline int min(int a, int b) {return a < b ? a : b;}
static inline int max(int a, int b) {return a > b ? a : b;}


int use_cooked = 1;
int use_multitap = 1;

void init_file_read(void)
{
	FILE *f;
	puts("loading edict");
	f = fopen("edict_sub_utf8_sorted.txt","r");
	fread(dict_all_lines, FILESIZE, 1, f);
	fclose(f);
	dict_all_lines_end[0] = '\n';
	dict_all_lines_end[1] = 0;
}


void init_dict_gen_index_lines(void)
{
	char *p,*q;
	int i;
	puts("locating lines");
	p = dict_all_lines;
	for(i = 0 ; i < FILELINES ; i++){
		dict_index_lines[i] = p;
		for(q = p ; *q != '\n' ; q++);
		*q = 0;
		p = q + 1;
	}
	dict_index_lines[i] = p;
}


void init_dict_gen_index_kanji_add(int line, wint_t kanji)
{
	indexhead_t *head = &dict_index_kanji[kanji];
	if(!head->array){
		head->size = 1;
		head->used = 0;
		head->array = malloc(head->size * sizeof(short int));
		//printf("%C %08d %08d %p\n", kanji, head->size, head->used, head->array);
	}else if(head->used >= head->size){
		head->size += 8;
		head->array = realloc(head->array, head->size * sizeof(short int));
		//printf("%C %08d %08d %p\n", kanji, head->size, head->used, head->array);
	}
	head->array[head->used++] = line;
}


void init_dict_gen_index_kanji_stats(void)
{
	wint_t i;
	int empty = 0;
	for(i = 0 ; i < MAXINDEX ; i++){
		indexhead_t *head = &dict_index_kanji[i];
		if(head->used != 0) printf("%08x %C %8d\n", i, i, head->used);
		else empty++;
	}
	printf("%8d empty heads\n", empty);
	printf("%8d used  heads\n", MAXINDEX - empty);
}


void init_dict_gen_index_kanji(void)
{
	int i,j,part;
	wchar_t vmin = WCHAR_MAX;
	wchar_t vmax = WCHAR_MIN;
	puts("indexing kanjis");
	for(i = 0 ; i < FILELINES ; i++){
		char const *p = dict_index_lines[i];
		wchar_t line[1024];

		//if(i < 18) continue;
		//if(i > 25) break;
		//printf("::\t%s\n", p);

		size_t n = mbsrtowcs(line, &p, 1023, NULL);

		part = 0;
		for(j = 0 ; j < n ; j++){
			wint_t c = line[j];
			switch(part){
			case 0:
				if(c == L'/'){ part = 4; break; }
				if(c == L' '){ part = 1; break; }
				vmin = min(c, vmin);
				vmax = max(c, vmax);
				init_dict_gen_index_kanji_add(i, c);
				//printf("0:\t%C\n", c);
				break;
			case 1:
				if(c == L'/'){ part = 4; break; }
				if(c == L'['){ part = 2; break; }
				//printf("1:\t%C\n", c);
				break;
			case 2:
				if(c == L'/'){ part = 4; break; }
				if(c == L']'){ part = 3; break; }
				vmin = min(c, vmin);
				vmax = max(c, vmax);
				init_dict_gen_index_kanji_add(i, c);
				//printf("2:\t%C\n", c);
				break;
			case 3:
				if(c == L'/'){ part = 4; break; }
				//printf("3:\t%C\n", c);
				break;
			case 4:
				if(c == L'/'){ part = 4; break; }
				init_dict_gen_index_kanji_add(i, c);
				//printf("4:\t%C\n", c);
				break;
			default:
				break;
			}
		}
	}

	printf("vmin: %d %d %C\n", MININDEX, vmin, vmin);
	printf("vmax: %d %d %C\n", MAXINDEX, vmax, vmax);
}


void init(void)
{
	init_file_read();
	init_dict_gen_index_lines();
	init_dict_gen_index_kanji();
#ifdef DEBUG
	//init_dict_gen_index_kanji_stats();
#endif
}


void dict_search(wchar_t const * const str, void (*fnc)(char const * const line))
{
	char mb[512];
	int i;

	wcstombs(mb, str, 511);

	for(i = 0 ; i < FILELINES ; i++){
		char const * const p = dict_index_lines[i];
		char const * const q = strstr(p, mb);
		if(q != NULL) (*fnc)(p);
	}
}


void out_printline(char const * const line)
{
	puts(line);
}


void out_backspace(void)
{
	printf("\e[1D");
}


void in_getstr_cooked(wchar_t *s, int maxlen)
{
	fgetws(s, maxlen, stdin);
}


void in_getstr_raw_startstop(int mode)
{
	static struct termios tio,tio2;

	switch(mode){
	case 0:
		tcgetattr(0, &tio);
		tio2 = tio;
		cfmakeraw(&tio2);
		tcsetattr(0, TCSAFLUSH, &tio2);
		break;
	case 1:
		tcsetattr(0, TCSAFLUSH, &tio);
		puts("");
		break;
	}
}


void in_getstr_multitap(wchar_t *s, int maxlen)
{
	static const wchar_t keypad[3][10][6] = {
	L"わをん。、0",
	L"まみむめも1",L"やゆよーー2",L"らりるれろ3",
	L"たちつてと4",L"なにぬねの5",L"はひふへほ6",
	L"あいうえお7",L"かきくけこ8",L"さしすせそ9",

	L"？！…：−0",
	L"まみむめも1",L"ゃゅょーー2",L"らりるれろ3",
	L"だぢづでど4",L"なにぬねの5",L"ばびぶべぼ6",
	L"ぁぃぅぇぉ7",L"がぎぐげご8",L"ざじずぜぞ9",

	L"ーーーーー0",
	L"まみむめも1",L"ゃゅょーー2",L"らりるれろ3",
	L"だぢづでど4",L"なにぬねの5",L"ぱぴぷぺぽ6",
	L"ぁぃぅぇぉ7",L"がぎぐげご8",L"ざじずぜぞ9",
	};

	wchar_t new, old, multi;
	int cnt,shift;

	in_getstr_raw_startstop(0);

restart:
	new = old = multi = -1, cnt = shift = 0;
	while(maxlen--){
		new = getwchar();
		if(new == '0') new = ',';
		else if(new == ',') new = '0';
		if(new == "BORRAR"){
			out_backspace();
			shift = 3;
			multi = L' ';
			goto restart;
		}
		if(multi > 0){
			if(new >= '0' && new <= '9'){
				if(new == old && !shift){
					cnt = (cnt + 1) % 6;
					multi = keypad[shift][old - '0'][cnt];
					out_backspace();
				}else{
					*s++ = multi;
				enter_multi:
					cnt = shift = 0;
					old = new;
					multi = keypad[shift][old - '0'][0];
				}
			}else if(new == ','){
				shift = (shift + 1) % 3;
				multi = keypad[shift][old - '0'][cnt];
				out_backspace();
			}else if(new == '\r' && old != '\r'){
				*s++ = multi;
				goto restart;
			}else if(new == '\r' && old == '\r'){
				break;
			}else{
				*s = 0;
				break;
			}
			putwchar(multi);
		}else{
			old = multi = -1; cnt = shift = 0;
			if(new == '\r'){ *s = 0; break; }
			if(new >= '0' && new <= '9') goto enter_multi;
			putwchar(new);
			*s++ = new;
		}
	}

	in_getstr_raw_startstop(1);
}


void in_getstr(wchar_t * s, int maxlen)
{
	wchar_t *t;
	int i;

	maxlen--;

	puts("\n\nenter string to search (keypad for multitap):");

	if(use_multitap){
		in_getstr_multitap(s, maxlen);
	}else if(use_cooked){
		in_getstr_cooked(s, maxlen);
	}

	for(t = s ;; t++){
		if(*t < ' '){
			*t = 0;
			break;
		}
	}

#ifdef DEBUG
	for(; *s ; s++){
		printf("%08x ", *s);
	}
	puts("\n");
#endif
}


int main(int argc, char **argv)
{
	setlocale(LC_CTYPE,"");
	init();

	for(;;){
		wchar_t str[512];
		in_getstr(str, 512);
		if(wcslen(str) == 0) break;
		dict_search(str, &out_printline);
	}

	return 0;
}
